from argparse import ArgumentParser
from importlib import import_module
from pydoc import locate
from yaml import YAMLError, safe_load


class BaseHandler():
    """
    Root parser and handler of arguments.
    """

    def __init__(self,
                 name='Default',
                 args: list[str] | None = None,
                 config_path: str | None = None,
                 handlers_dir='handlers') -> None:
        self.__name = name
        self.__handlers_dir = handlers_dir
        self.__register_name()
        config = self._load_config(config_path)
        default = config.pop('default', {})
        self.__default_group = default.get('group', {})
        self.__default_command = default.get('command', {})
        self.__default_argument = default.get('argument', {})
        handlers_config = config.pop('handlers', {})
        self.__handlers_root = handlers_config.get('root', 'default')
        self.__handlers_internal = handlers_config.get('internal', True)
        self._args = self.__parse(config, args)

    @classmethod
    def meta(cls, name='Default'):
        if (not hasattr(BaseHandler, '_meta')) or (name not in cls._meta):
            raise ValueError(
                f"BaseHandler with name '{name}' is not defined"
                "(${cls._meta.get(name)})")
        args = cls._meta[name]._args

        class CustomHandler:
            @property
            def args(self):
                return args
        return CustomHandler

    def __register_name(self) -> None:
        if not hasattr(BaseHandler, '_meta'):
            BaseHandler._meta: dict[str, BaseHandler] = {}
        if self.__name in BaseHandler._meta:
            raise ValueError(
                f"BaseHandler with name '{self.__name}' is already defined")
        BaseHandler._meta[self.__name] = self

    @staticmethod
    def _load_config(config_path: str | None) -> dict:
        if config_path is None:
            config_path = 'cli-config.yaml'
        with open(config_path, 'r', encoding='UTF-8') as stream:
            try:
                config = safe_load(stream)
                if config is None:
                    raise TypeError(f"Config '{config_path}' cannot be empty")
                if not isinstance(config, dict):
                    raise TypeError(
                        f"Invalid config '{config_path}': " +
                        f"{dict} expected, but {type(config)} found")
                return config
            except YAMLError as error:
                print(error)
                exit(1)

    def __parse(self, config: dict, args: list[str] | None) -> dict:
        group = config.pop('group', {})
        rules = config.pop('next', [])
        parser = ArgumentParser(**config)
        self.__next(parser, rules, group)
        namespace = parser.parse_args(args)
        return vars(namespace)

    def __next(self,
               parser: ArgumentParser,
               rules: list[dict],
               group: dict[str, str],
               dest: str | None = None) -> None:
        subparsers = None
        for rule in rules:
            if 'command' in rule:
                if subparsers is None:
                    if dest is None:
                        dest = self.__default_group.get('command', 'command')
                    options = {"dest": dest, "title": 'commands',
                               "required": True, **self.__default_group, **group}
                    subparsers = parser.add_subparsers(**options)
                command = rule.pop('command')
                new_next = rule.pop('next', None)
                new_group = rule.pop('group', {})
                new_parser = subparsers.add_parser(
                    command, **self.__default_command, **rule)
                if new_next is not None:
                    self.__next(new_parser, new_next, new_group)
            elif 'argument' in rule:
                argument = rule.pop('argument')
                options = {**self.__default_argument, **rule}
                if 'action' in rule and 'type' in self.__default_argument:
                    options.pop('type', None)
                if 'type' in options:
                    argtype = locate(options.pop('type'))
                    if argtype is not None:
                        options['type'] = argtype
                parser.add_argument(*argument, **options)
            else:
                raise ValueError(f"Invalid rule: {rule}")

    def __get_path(self, handlers_dir: str) -> str:
        tree = []
        level = 'command'
        if self._args.get(level):
            while level := self._args.get(level):
                tree.append(level)
        else:
            if self.__handlers_root:
                tree.append(self.__handlers_root)
        prefix = '._' if self.__handlers_internal else '.'
        if len(tree) == 0:
            return handlers_dir
        return handlers_dir + prefix + '_'.join(tree)

    def run(self):
        """
        Run the handler.
        """
        path = self.__get_path(self.__handlers_dir)
        import_module(path).Handler()

    @property
    def args(self) -> dict:
        """
        Get the arguments.
        """
        return self._args



from argyaml import BaseHandler
from yaml import safe_load
import os


class Handler(BaseHandler.meta('Default')):
    def __init__(self):
        config_path = self.args.get('config')
        if not config_path:
            config_path = 'cli-config.yaml'
        config = BaseHandler._load_config(config_path)
        rules = config.get('next', [])
        handlers_config = config.get('handlers', {})
        handlers_root = handlers_config.get('root', 'default')
        handlers_internal = handlers_config.get('internal', True)
        handler_name = self.args.get('name')
        handler_meta = f"'{handler_name}'" if handler_name else ''
        handlers_dir = self.args.get('dir') or 'handlers'
        print(f'Generating handlers into {handlers_dir}/')
        for file in self.__get_handlers('', rules):
            if not file:
                file = handlers_root
            if not handlers_internal:
                file = file[1:]
            try:
                if not os.path.exists(handlers_dir):
                    os.makedirs(handlers_dir)
                flag = 'w' if self.args.get('force', False) else 'x'
                with open(f'{handlers_dir}/{file}.py', flag,
                          encoding='UTF-8') as stream:
                    stream.writelines(
                        f"from argyaml import BaseHandler\n\n\n"
                        f"class Handler(BaseHandler.meta({handler_meta})):\n"
                        f"    def __init__(self):\n"
                        f"        pass\n")
                    print('created:', f'{file}.py')
            except FileExistsError:
                print('skipped:', f'{file}.py')

    def __get_handlers(self, base: str, rules: list):
        subcommand = False
        for rule in rules:
            if 'command' in rule:
                subcommand = True
                subrules = rule.get('next', [])
                yield from self.__get_handlers(
                    base + '_' + rule['command'], subrules)
        if not subcommand:
            yield base



from argyaml import BaseHandler


def run(command: str | None = None) -> None:
    """
    Run argyaml cli tool
    """
    import os
    from sys import argv
    BaseHandler(
        args=[command] + argv[1:] if command else None,
        config_path=os.path.join(os.path.dirname(__file__), 'cli-config.yaml'),
        handlers_dir='argyaml.handlers').run()


def init():
    raise NotImplementedError('Init script is not implemented')
    run('init')


def generate():
    run('generate')


if __name__ == '__main__':
    run()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['argyaml', 'argyaml.handlers']

package_data = \
{'': ['*']}

install_requires = \
['pyyaml>=6.0,<7.0']

entry_points = \
{'console_scripts': ['generate = argyaml.__main__:generate',
                     'main = argyaml.__main__:run']}

setup_kwargs = {
    'name': 'argyaml',
    'version': '0.1.1',
    'description': 'Create a powerful and efficient CLI application using simple and user-friendly yaml config',
    'long_description': '# Argyaml\nArgyaml is a small module for creating _powerful_ and _scalable_ __CLI applications__ based on a _simple_ and _user-friendly_ yaml __configuration file__.\n\n### Motivation\nArgyaml is built over the [argparse](https://docs.python.org/3/library/argparse.html) module, which is a part of python standard library starting python 3.2. While it works well for tiny projects that need to quickly access a few arguments and provide automatically generated help and usage messages for user, it gets very complicated and painful when it comes to large projects or your application grows in complexity.\n\n### Features\n- Independent specification of CLI commands and arguments.\n- No boilerplate code.\n- Ability to set default options for commands, groups, and arguments.\n- Automatic and optimized invocation of command handlers.\n- Handler template files generator.\n\n## Install\n\n```bash\n# pip\npip install argyaml\n\n# poetry\npoetry add argyaml\n```\n\n## Getting started\n```yaml\n# cli-config.yaml\nprog: todo\ndescription: My beautiful todo app\nnext:\n  - command: new\n    next:\n      - command: task\n        description: Create a new task\n        next:\n          - argument: [name]\n            help: the name of task\n  - command: list\n    next:\n      - argument: [-t, --task]\n        help: display tasks only\n        action: \'store_true\'\n```\n\n```python\nfrom argyaml import BaseHandler\n\nbase = BaseHandler()\nbase.args # <-- parsed and ready-to-use arguments\n```\n\n__Learn more about [config file](./wiki#Config).__\n\n\n### Using Handlers\nGenerate template files using `argyaml generate`:\n```bash\n# pip\npython -m argyaml generate\n\n# poetry\npoetry run python -m argyaml generate\n```\nThis will generate the following files:\n```new\nhandlers/\n  _new_task.py\n  _list.py\n```\nNow, whenever `new task` command is called, the corresponding handler init function will be invoked with all additional aruments stored in the `self.args` object.\n\n```python\n# _new_task.py\nfrom argyaml import BaseHandler\n\nclass Handler(BaseHandler.meta()):\n    def __init__(self):\n        print(f"Successfully created task \'{self.args[\'name\']}\'!")\n```\nModify the main file to run the base handler:\n```python\nfrom argyaml import BaseHandler\n\nbase = BaseHandler()\nbase.run()\n```\n\n__Learn more about [BaseHandler](./wiki#BaseHandler) and [argyaml generator](./wiki#Generator).__\n\n## Contributing\nFeel free to open issues. Pull requests are welcome.\n\n## License\nThis project is licensed under the [MIT License](./LICENSE).\n',
    'author': 'Artur Sharapov',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
